@@ -63,7 +63,7 @@
 #include <libkmod.h>
 #include "xf86drm.h"
 #include "xf86drmMode.h"
-#include "json.h"
+#include "json-parser/json.h"
 
 static inline void freep(void *);
 static inline void fclosep(FILE **);
@@ -83,18 +83,16 @@ static inline void pclosep(FILE **);
 #define LAST_BOOT "/var/lib/ubuntu-drivers-common/last_gfx_boot"
 #define OFFLOADING_CONF "/var/lib/ubuntu-drivers-common/requires_offloading"
 #define RUNTIMEPM_OVERRIDE "/etc/u-d-c-nvidia-runtimepm-override"
+#define UDEV_NVIDIA_COMPLETED "/run/u-d-c-nvidia-drm-was-loaded"
 #define XORG_CONF "/etc/X11/xorg.conf"
 #define KERN_PARAM "nogpumanager"
 #define AMDGPU_PRO_PX  "/opt/amdgpu-pro/bin/amdgpu-pro-px"
+#define CHASSIS_PATH "/sys/devices/virtual/dmi/id/chassis_type"
 
 #define AMD 0x1002
 #define INTEL 0x8086
 #define NVIDIA 0x10de
 
-#ifndef PCI_CAP_ID_NULL
-#define PCI_CAP_ID_NULL 0x0
-#endif
-
 #define MAX_CARDS_N 10
 
 typedef enum {
@@ -1014,40 +1012,124 @@ void find_disabled_cards(char *dir, struct device **devices,
 }
 
 
-/* Check if a kernel module is available for the current kernel */
+static int nvidia_dir_filter(const struct dirent *entry)
+{
+    return (strncmp(entry->d_name, "nvidia", 6) == 0);
+}
+
+
+static int nvidia_module_filter(const struct dirent *entry)
+{
+    return (strncmp(entry->d_name, "nvidia.ko", 9) == 0);
+}
+
+
+static int amdgpu_dir_filter(const struct dirent *entry)
+{
+    return (strncmp(entry->d_name, "amdgpu", 6) == 0);
+}
+
+
+static int amdgpu_module_filter(const struct dirent *entry)
+{
+    return (strncmp(entry->d_name, "amdgpu.ko", 9) == 0);
+}
+
+
+/* Go through a directory, looking for a kernel module
+ *
+ */
+static int search_dir_for_module(const char *kernel_path, int (*dir_filter)(const struct dirent *),
+                                 int (*mod_filter)(const struct dirent *), const char *module)
+{
+    int status = 0;
+    int n = 0, m = 0;
+    struct dirent **dirlist;
+    char current_path[PATH_MAX];
+    char module_name[strlen(module)+4];
+
+    snprintf(module_name, sizeof(module_name), "%s.ko", module);
+    n = scandir(kernel_path, &dirlist, dir_filter, alphasort);
+    if (n > 0) {
+        int i = n;
+        /* Check the nvidia version from high to low */
+        while (i--) {
+            snprintf(current_path, sizeof(current_path), "%s/%s", kernel_path, dirlist[i]->d_name);
+            if (strstr(dirlist[i]->d_name, module_name) != NULL) {
+                status = 1;
+                fprintf(log_handle, "Found %s module in %s\n", module_name, current_path);
+                break;
+            }
+            else {
+                fprintf(log_handle, "Looking for %s modules in %s\n", module, current_path);
+                m = search_dir_for_module(current_path, mod_filter, dir_filter, module);
+                if (m > 0) {
+                    status = 1;
+                    break;
+                }
+            }
+        }
+        for (i = 0; i < n; i++) {
+            if (dirlist[i] != NULL)
+                free(dirlist[i]);
+        }
+        if (dirlist != NULL)
+            free(dirlist);
+        if (status)
+            return status;
+    }
+    return status;
+}
+
+
+/* Check if a kernel module is available for the current kernel
+ *
+ * NOTE: This is only meant to check extra modules such as
+ *       nvidia, or amdgpu pro.
+ */
 static bool is_module_available(const char *module)
 {
-    char dir[PATH_MAX];
-    struct dirent *dp;
-    DIR *dfd;
+    int (*dir_filter)(const struct dirent *);
+    int (*mod_filter)(const struct dirent *);
+    char kernel_path[PATH_MAX];
     struct utsname uname_data;
-    bool status = false;
+    int status = 0;
+
+    if (strcmp(module, "nvidia") == 0) {
+        dir_filter = nvidia_dir_filter;
+        mod_filter = nvidia_module_filter;
+    }
+    else if (strcmp(module, "amdgpu") == 0) {
+        dir_filter = amdgpu_dir_filter;
+        mod_filter = amdgpu_module_filter;
+    }
+    else {
+        fprintf(log_handle, "Checking for the %s module avilability is not supported\n", module);
+        return false;
+    }
 
     if (uname(&uname_data) < 0) {
         fprintf(stderr, "Error: uname failed\n");
         return false;
     }
 
-    sprintf(dir, "/lib/modules/%s/updates/dkms", uname_data.release);
+    sprintf(kernel_path, "/lib/modules/%s/kernel", uname_data.release);
+    fprintf(log_handle, "Looking for %s modules in %s\n", module, kernel_path);
 
-    fprintf(log_handle, "Looking for %s modules in %s\n", module, dir);
+    /* Look for the linux-restricted-modules first */
+    status = search_dir_for_module(kernel_path, dir_filter, mod_filter, module);
 
-    if ((dfd = opendir(dir)) == NULL) {
-        fprintf(stderr, "Error: can't open %s\n", dir);
-        return false;
+    if (status > 0) {
+        return true;
     }
+    else {
+        /* Look for dkms modules */
+        sprintf(kernel_path, "/lib/modules/%s/updates/dkms", uname_data.release);
+        fprintf(log_handle, "Looking for %s modules in %s\n", module, kernel_path);
+        status = search_dir_for_module(kernel_path, dir_filter, mod_filter, module);
 
-    while ((dp = readdir(dfd)) != NULL) {
-        if (!starts_with(dp->d_name, module))
-            continue;
-
-        status = true;
-        fprintf(log_handle, "Found %s module: %s\n", module, dp->d_name);
-        break;
+        return (status > 0);
     }
-    closedir(dfd);
-
-    return status;
 }
 
 
@@ -1962,6 +2044,7 @@ static bool enable_prime(const char *prime_settings,
                         const struct device *device,
                         struct device **devices,
                         int cards_n) {
+    int major, minor, extra;
     bool status = false;
     int tries = 0;
     /* Check if prime_settings is available
@@ -1981,6 +2064,7 @@ static bool enable_prime(const char *prime_settings,
 
     get_prime_action();
     if (prime_mode == ON) {
+prime_on_fallback:
         /* Create an OutputClass just for PRIME, to override
          * the default NVIDIA settings
          */
@@ -1992,6 +2076,16 @@ static bool enable_prime(const char *prime_settings,
             load_module("nvidia");
     }
     else if (prime_mode == ONDEMAND) {
+        if (get_nvidia_driver_version(&major, &minor, &extra)) {
+            if (major < 450) {
+                /* Set prime to "on" */
+                fprintf(log_handle, "Info: falling back to on mode for PRIME, since driver series %d < 450.\n",
+                major);
+                create_prime_settings(prime_settings);
+                goto prime_on_fallback;
+            }
+        }
+
         /* Create the ServerLayout required to enabling offload
          * for NVIDIA.
          */
@@ -2300,6 +2394,89 @@ static bool is_nv_runtimepm_enabled(struct pci_dev *device) {
     return false;
 }
 
+static bool is_laptop(void) {
+    /*
+        Chassis types:
+
+        01 Other
+        02 Unknown
+        03 Desktop
+        04 Low Profile Desktop
+        05 Pizza Box
+        06 Mini Tower
+        07 Tower
+        08 Portable
+        09 Laptop
+        10 Notebook
+        11 Hand Held
+        12 Docking Station
+        13 All In One
+        14 Sub Notebook
+        15 Space-saving
+        16 Lunch Box
+        17 Main Server Chassis
+        18 Expansion Chassis
+        19 Sub Chassis
+        20 Bus Expansion Chassis
+        21 Peripheral Chassis
+        22 RAID Chassis
+        23 Rack Mount Chassis
+        24 Sealed-case PC
+        25 Multi-system
+        26 CompactPCI
+        27 AdvancedTCA
+        28 Blade
+        29 Blade Enclosing
+        30 Tablet
+        31 Convertible
+        32 Detachable
+        33 IoT Gateway
+        34 Embedded PC
+        35 Mini PC
+        36 Stick PC
+    */
+
+    _cleanup_free_ char *line = NULL;
+    _cleanup_fclose_ FILE *file = NULL;
+    size_t len = 0;
+    size_t read;
+    int code = 0;
+    int status;
+    bool is_laptop = false;
+
+    file = fopen(CHASSIS_PATH, "r");
+    if (file == NULL) {
+        fprintf(log_handle, "Error: can't open %s\n", CHASSIS_PATH);
+        return false;
+    }
+    else {
+        while ((read = getline(&line, &len, file)) != -1) {
+            status = sscanf(line, "%d\n", &code);
+            if (status) {
+                fprintf(log_handle, "Chassis type: \"%d\"\n", code);
+
+                switch (code)
+                {
+                    case 8:
+                    case 9:
+                    case 10:
+                    case 31:
+                        fprintf(log_handle,"Laptop detected\n");
+                        is_laptop = true;
+                        break;
+
+                    default:
+                        fprintf(log_handle,"Laptop not detected\n");
+                        break;
+                }
+
+                break;
+            }
+        }
+    }
+    return is_laptop;
+}
+
 
 int main(int argc, char *argv[]) {
 
@@ -2352,6 +2529,8 @@ int main(int argc, char *argv[]) {
     /* Device information (discrete) */
     struct device discrete_device = { 0 };
 
+    /* Get nvidia driver version */
+    int major, minor, extra;
 
     while (1) {
         static struct option long_options[] =
@@ -2750,12 +2929,18 @@ int main(int argc, char *argv[]) {
                         has_nvidia = true;
 
                         if (!current_devices[cards_n]->boot_vga) {
-                            nvidia_runtimepm_supported = is_nv_runtimepm_supported(dev->device_id);
+                            /* Do not enable RTD3 unless it's a laptop */
+                            if (is_laptop()) {
+                                nvidia_runtimepm_supported = is_nv_runtimepm_supported(dev->device_id);
 
-                            if (!nvidia_runtimepm_supported) {
-                                /* This is a fairly expensive call, so check the database first */
-                                get_d3_substates(pm_dev, &d3cold, &d3hot);
-                                nvidia_runtimepm_supported = d3hot;
+                                if (!nvidia_runtimepm_supported) {
+                                    /* This is a fairly expensive call, so check the database first */
+                                    get_d3_substates(pm_dev, &d3cold, &d3hot);
+                                    nvidia_runtimepm_supported = d3hot;
+                                }
+                            }
+                            else {
+                                nvidia_runtimepm_supported = false;
                             }
 
                             fprintf(log_handle, "Is nvidia runtime pm supported for \"0x%x\"? %s\n", dev->device_id,
@@ -2803,6 +2988,34 @@ int main(int argc, char *argv[]) {
         pci_cleanup(pacc);
         pacc = NULL;
     }
+
+    /* Probe graphic drivers earlier since some graphic related application
+     * (e.g. gdm, mutter) rely on udev rules to apply extra configurations.
+     */
+    if (has_nvidia && !nvidia_blacklisted && !is_module_loaded("nvidia"))
+        load_module("nvidia");
+
+    /* Make sure the udev rules are applied for graphic cards.
+     *
+     * Also make sure to do this before trying to open /dev/dri/card*
+     * because doing so too early may trigger a nvidia bug:
+     * https://forums.developer.nvidia.com/t/545-29-06-18-1-flip-event-timeout-error-on-startup-shutdown-and-sometimes-suspend-wayland-unusable/274788
+     */
+    if (has_nvidia && !nvidia_blacklisted) {
+        for (i = 0; i < 500; i++, usleep(20000)) {
+            if (is_file(UDEV_NVIDIA_COMPLETED))
+                break;
+            /* Get nvidia driver version inside loop because sometime nvidia
+             * driver take longer time to load.
+             */
+            if (get_nvidia_driver_version(&major, &minor, &extra) &&
+                major < 390) {
+                break;
+            }
+        }
+        fprintf(log_handle, "Takes %dms to wait for nvidia udev rules completed.\n", i * 20);
+    }
+
     /* Add information about connected outputs */
     add_connected_outputs_info(current_devices, cards_n);
 
